"""Numpy random seeder"""

import numpy as np


def random_seed(seed):
    seed = int(seed)
    if seed < 0 or seed > (2 ** 32 - 1):
        raise ValueError("Seed must be between 0 and 2**32 - 1")
    np.random.seed(seed)
